/*
geotool.c - some GeoIP.dat quirks

Copyright (c) 2006 NoisyB


This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#ifdef  HAVE_CONFIG_H
#include "config.h"
#endif
#ifdef  USE_GEOIP
#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <time.h>
#include <GeoIP.h>
#include "misc/string.h"
#include "misc/getopt.h"
#include "misc/getopt2.h"
#include "misc/misc.h"
#include "geotool_defines.h"
#include "geotool.h"


st_geotool_t geotool;


static const char *
u32_to_ip (unsigned long i)
{
  static char buf[32];

  sprintf (buf, "%d.%d.%d.%d", (int) (i >> 24) & 0xff,
                             (int) (i >> 16) & 0xff,
                             (int) (i >> 8) & 0xff,
                             (int) i & 0xff);
  return buf;
}


static int
geotrace_isaddr (const char *a)
{
  for (; *a; a++)
    if (isalpha (*a))
      return 0;
  return 1;
}


static const char *
geotrace_get_country (const char *a)
{
  int pos = 0;

  if (geotrace_isaddr (a))
    pos = GeoIP_id_by_addr (geotool.gi, a);
  else
    pos = GeoIP_id_by_name (geotool.gi, a);

  return GeoIP_country_name[pos];
}


static void
geotrace_traceroute_pipe (FILE *in, FILE *out)
{
  char buf[MAXBUFSIZE];
  char tmp[MAXBUFSIZE];
  char addr[MAXBUFSIZE];
  char *p = NULL;

  while (fgets (buf, MAXBUFSIZE, in))
    {
      p = (char *) memmem2 (buf, strlen (buf), "(x.", 3, MEMCMP2_WCARD ('x'));
      if (!p)
        p = (char *) memmem2 (buf, strlen (buf), "(xx.", 4, MEMCMP2_WCARD ('x'));
      if (!p)
        p = (char *) memmem2 (buf, strlen (buf), "(xxx.", 5, MEMCMP2_WCARD ('x'));

      if (p)
        {
          strcpy (addr, p + 1);
          *(strchr (addr, ')')) = 0;

          sprintf (tmp, "%s, %s", addr, geotrace_get_country (addr));

#ifdef  DEBUG
              printf ("%s\n", tmp);
              fflush (stdout);
#endif

          strins (p + 1, strchr (p, ')') - (p + 1), tmp);
        }

      fprintf (out, buf);
    }
}


static int
geotool_guarding_p2p (st_geotool_t *gt, int skip)
{
  FILE *out = NULL;
  unsigned long i = 0;
  const char *p = NULL;
  unsigned long range_start = 0;
  char range_country[5];
  unsigned long start = 0;
  time_t t1 = 0, t2 = 0;

  if (!(out = fopen ("guarding.p2p", "w")))
    {
      fprintf (stderr, "ERROR: could not open \"guarding.p2p\" for writing\n");
      return -1;
    }

  t1 = t2 = time (0);
  range_start = start;
  strcpy (range_country, "?");

  for (i = start; i <= 0xffffffff; i += skip)
    {
      p = GeoIP_country_name_by_addr (gt->gi, u32_to_ip (i));
  
      if (!p)
        p = "?";

      if (strcmp (range_country, p))
        {
          fprintf (out, "%s:%s-", range_country, u32_to_ip (range_start));
          fprintf (out, "%s\n", u32_to_ip (i - 1));
          fflush (out);

          range_start = i;
          strcpy (range_country, p);
        }

      if (t2 < t1 - 1)
//      if (!(i % 0x1000000))
        {
          printf ("Progress: %d%% (%s, %ld, %s)%*s\r", misc_percent (i, 0xffffffff), u32_to_ip (i), i, p, 20, " ");
          fflush (stdout);
          t2 = t1;
        }

      t1 = time (0);
    }

  fclose (out);

  return 0;
}

#endif  // USE_GEOIP

int
main (int argc, char **argv)
{
#ifndef  USE_GEOIP
  (void) argc;
  (void) argv;
#else
  static char geoip_dat[FILENAME_MAX];
  char short_options[ARGS_MAX];
  struct option long_only_options[ARGS_MAX];
  int option_index = 0;
  int c = 0;
  int value = 0;
  const st_getopt2_t options[] = {
    {
      NULL,      0, 0, 0, NULL,
      "geotool " GEOTOOL_VERSION_S " " CURRENT_OS_S " 2006 by NoisyB\n"
      "This may be freely redistributed under the terms of the GNU Public License\n\n"
      "Usage: geotool [OPTION]... GeoIP.dat\n", NULL
    },
    {
      "trace",    1, 0, GEOTOOL_TRACE,
      "IP",   "show geological position of IP or domain\n"
              OPTION_LONG_S "trace=- will read (for example) traceroute output\n"
              "from stdin and insert geographic information behind\n"
              "IP's or domains",
              NULL
    },
#if 0
    {
      "html",    0, 0, GEOTOOL_HTML,
      NULL,   "output as html with map links to stdout", NULL
    },
    {
      "rss",    0, 0, GEOTOOL_RSS,
      NULL,   "output as RSS with map links to stdout", NULL
    },
#endif
    {
      "reverse",    2, 0, GEOTOOL_REVERSE,
      "SKIP",  "reverse GeoIP.dat into a list of countries with their\n"
               "IP ranges (guarding.p2p); SKIP sets how many IP's\n"
               "should be skipped before the next check\n"
               "0     will result in the most correct list\n"
               "256   for slow computers (default)\n"
               "65536 fastest", NULL
    },
    {
      "version", 0, 0, GEOTOOL_VERSION,
      NULL,   "output version information and exit", NULL
    },
    {
      "ver",     0, 0, GEOTOOL_VERSION,
      NULL,   NULL, NULL
    },
    {
      "help",    0, 0, GEOTOOL_HELP,
      NULL,   "display this help and exit", NULL
    },
    {
      "h",       0, 0, GEOTOOL_HELP,
      NULL,   NULL, NULL
    },
    {
      NULL,      0, 0, 0,
      NULL, "\nReport problems/ideas/whinge to noisyb@gmx.net\n", NULL},
    {NULL,      0, 0, 0,   NULL,   NULL, NULL}
  };

  if (argc < 2)
    {
      getopt2_usage (options);
      return 0;
    }

  memset (&geotool, 0, sizeof (st_geotool_t));

  strncpy (geoip_dat, argv[argc - 1], FILENAME_MAX)[FILENAME_MAX - 1] = 0;
  if (access (geoip_dat, F_OK) != 0)
    strcpy (geoip_dat, GEOTOOL_GEOIP_DATABASE);

  if (!(geotool.gi = GeoIP_open (geoip_dat, GEOIP_STANDARD)))
    {
      fprintf (stderr, "ERROR: could not open \"%s\"\n", geoip_dat);
      getopt2_usage (options);
      return -1;
    }

  getopt2_short (short_options, options, ARGS_MAX);
  getopt2_long_only (long_only_options, options, ARGS_MAX);

  while ((c = getopt_long_only (argc, argv, short_options, long_only_options, &option_index)) != -1)
    switch (c)
      {
        case GEOTOOL_HTML:
          geotool.html = 1;
          break;

        case GEOTOOL_RSS:
          geotool.rss = 1;
          break;

        case GEOTOOL_TRACE:
          if (!strcmp (optarg, "-"))
            geotrace_traceroute_pipe (stdin, stdout);
          else
            printf ("%s\n", geotrace_get_country (optarg));
          break;

        case GEOTOOL_REVERSE:
          value = 0x100; // default
          if (optarg)
            value = strtol (optarg, NULL, 10);
          geotool_guarding_p2p (&geotool, value > 0 ? value : 1);
          break;

        case GEOTOOL_HELP:
          getopt2_usage (options);
          return 0;

        case GEOTOOL_VERSION:
          printf ("%s\n", GEOTOOL_VERSION_S);
          return 0;

        default:
          printf ("Try '%s --help' for more information.\n", argv[0]);
          return -1;
      }

  GeoIP_delete (geotool.gi);

#endif  // USE_GEOIP
  return 0;
}
